// -*- c++ -*-

/*
    Copyright (C) 2009  David Psenicka
    This file is part of FOMUS.

    FOMUS is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FOMUS is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef FOMUS_FOMUSAPI_H
#define FOMUS_FOMUSAPI_H

#include "basetypes.h"

#ifdef __cplusplus
extern "C" {
#endif

  #define FOMUS_API_VERSION 0
  
  // ENUMS
  enum fomus_param {
    fomus_par_none,
    fomus_par_entry, // 1
    fomus_par_list, // 2
    fomus_par_setting,
    fomus_par_settingval, // 3, 4
    fomus_par_locfile,
    fomus_par_locline,
    fomus_par_loccol, // 5 - 7

    // instruments and parts
    fomus_par_import,
    fomus_par_import_settingval,

    fomus_par_export,
    fomus_par_export_settingval,

    fomus_par_clef,
    fomus_par_clef_settingval,

    fomus_par_staff,
    fomus_par_staff_clefs, 
    fomus_par_staff_settingval,

    fomus_par_percinst,
    fomus_par_percinst_template,
    fomus_par_percinst_id, // 19
    fomus_par_percinst_imports,
    fomus_par_percinst_export,
    fomus_par_percinst_settingval,

    fomus_par_inst,
    fomus_par_inst_template,
    fomus_par_inst_id, // 27
    fomus_par_inst_staves,
    fomus_par_inst_imports,
    fomus_par_inst_export,
    fomus_par_inst_percinsts,
    fomus_par_inst_settingval,

    fomus_par_part, // 35
    fomus_par_part_id,
    fomus_par_part_inst, // 39
    fomus_par_part_settingval,

    fomus_par_partmap,
    fomus_par_partmap_part,
    fomus_par_partmap_metapart,
    fomus_par_partmap_settingval,

    fomus_par_metapart,
    fomus_par_metapart_id,
    fomus_par_metapart_partmaps,
    fomus_par_metapart_settingval,
    
    fomus_par_measdef,
    fomus_par_measdef_id,
    fomus_par_measdef_settingval,

    fomus_par_list_percinsts,
    fomus_par_list_insts, // 56
    fomus_par_settingval_percinsts,
    fomus_par_settingval_insts, // 58
    
    // REGIONS
    fomus_par_region, // set a region id (0 == no ID)
    fomus_par_region_voice,
    fomus_par_region_voicelist,
    fomus_par_region_time,
    fomus_par_region_gracetime,
    fomus_par_region_duration,
    fomus_par_region_pitch,
    fomus_par_region_dynlevel,
    fomus_par_region_mark,
    fomus_par_region_settingval,
    
    // EVENTS
    fomus_par_time, // 68
    fomus_par_gracetime, // gracetime resets to 'none' every time off is changed in any way
    fomus_par_duration,
    fomus_par_pitch,
    fomus_par_dynlevel,
    fomus_par_voice,

    fomus_par_note_settingval,

    fomus_par_meas, // 75
    fomus_par_meas_measdef,
    
    // marks
    fomus_par_markid, // 77
    fomus_par_markval,
    fomus_par_mark, // 79

    // events
    fomus_par_noteevent, // 80
    fomus_par_restevent,
    fomus_par_markevent,

    fomus_par_events,
    
    fomus_par_n
  };

  enum fomus_action {
    fomus_act_none,
    
    fomus_act_set, // (used most often)
    
    fomus_act_inc,
    fomus_act_dec,
    fomus_act_mult,
    fomus_act_div,

    fomus_act_start, // 6
    fomus_act_append, // for appending a list (instead of `set')
    fomus_act_add, // 8
    fomus_act_remove,
    fomus_act_end, // 10
    fomus_act_clear, // for marks, which are more complicated--also reset to default

    fomus_act_queue,
    fomus_act_cancel,
    fomus_act_resume,

    fomus_act_n
  };

  // STDOUT and STDERR STREAM via CALLBACK FUNCTION
  typedef void (*fomus_output)(const char* str);
  
#ifndef LIBFOMUS_HIDE

  LIBFOMUS_EXPORT int fomus_api_version();
  
  LIBFOMUS_EXPORT extern int fomus_err; // TRUE if there's an error!

  LIBFOMUS_EXPORT const char* fomus_version();
    
  // data
  LIBFOMUS_EXPORT void fomus_init(); // MUST call before anything, can be called repeatedly at any time, frees and invalidates all instances, if err, than can't continue
  LIBFOMUS_EXPORT FOMUS fomus_new(); // new instance
  LIBFOMUS_EXPORT void fomus_free(FOMUS f); // destroy instance

  LIBFOMUS_EXPORT void fomus_clear(FOMUS f); // clear all input 

  // if f is NULL, you get the default value (only need to do this from within validdeps function)
  LIBFOMUS_EXPORT fomus_int fomus_get_ival(FOMUS f, const char* set);
  LIBFOMUS_EXPORT struct fomus_rat fomus_get_rval(FOMUS f, const char* set);
  LIBFOMUS_EXPORT fomus_float fomus_get_fval(FOMUS f, const char* set);
  LIBFOMUS_EXPORT const char* fomus_get_sval(FOMUS f, const char* set);
// #define GET_SETF fomus_get_fval
// #define GET_SETR fomus_get_rval
// #define GET_SETI fomus_get_ival
// #define GET_SETS fomus_get_sval
  
  // VALUES
  //void fomus_bval(FOMUS f, enum fomus_param par, enum fomus_action act, fomus_bool val);
//   LIBFOMUS_EXPORT void fomus_ival(FOMUS f, enum fomus_param par, enum fomus_action act, fomus_int val);
//   LIBFOMUS_EXPORT void fomus_rval(FOMUS f, enum fomus_param par, enum fomus_action act, fomus_int num, fomus_int den);
//   LIBFOMUS_EXPORT void fomus_mval(FOMUS f, enum fomus_param par, enum fomus_action act, fomus_int val, fomus_int num, fomus_int den);
//   LIBFOMUS_EXPORT void fomus_fval(FOMUS f, enum fomus_param par, enum fomus_action act, fomus_float val);
//   LIBFOMUS_EXPORT void fomus_sval(FOMUS f, enum fomus_param par, enum fomus_action act, const char* str);
//   LIBFOMUS_EXPORT void fomus_act(FOMUS f, enum fomus_param par, enum fomus_action act); // use fomus_append action to append (or fomus_set to replace)

  LIBFOMUS_EXPORT void fomus_ival(FOMUS f, int par, int act, fomus_int val);
  LIBFOMUS_EXPORT void fomus_rval(FOMUS f, int par, int act, fomus_int num, fomus_int den);
  LIBFOMUS_EXPORT void fomus_mval(FOMUS f, int par, int act, fomus_int val, fomus_int num, fomus_int den);
  LIBFOMUS_EXPORT void fomus_fval(FOMUS f, int par, int act, fomus_float val);
  LIBFOMUS_EXPORT void fomus_sval(FOMUS f, int par, int act, const char* str);
  LIBFOMUS_EXPORT void fomus_act(FOMUS f, int par, int act); // use fomus_append action to append (or fomus_set to replace)

  LIBFOMUS_EXPORT void fomus_rt();
  
  LIBFOMUS_EXPORT void fomus_ivalrt(FOMUS f, int par, int act, fomus_int val);
  LIBFOMUS_EXPORT void fomus_rvalrt(FOMUS f, int par, int act, fomus_int num, fomus_int den);
  LIBFOMUS_EXPORT void fomus_mvalrt(FOMUS f, int par, int act, fomus_int val, fomus_int num, fomus_int den);
  LIBFOMUS_EXPORT void fomus_fvalrt(FOMUS f, int par, int act, fomus_float val);
  LIBFOMUS_EXPORT void fomus_svalrt(FOMUS f, int par, int act, const char* str);
  LIBFOMUS_EXPORT void fomus_actrt(FOMUS f, int par, int act); // use fomus_append action to append (or fomus_set to replace)

  // LOAD UP A FILE
  LIBFOMUS_EXPORT void fomus_load(FOMUS f, const char* filename);
  LIBFOMUS_EXPORT void fomus_parse(FOMUS f, const char* input);
  
  LIBFOMUS_EXPORT void fomus_run(FOMUS f); // INVALIDATES FOMUS INSTANCE, EVEN IF THERE's AN ERROR!  If you want to reuse instance, then copy it with fomus_copy first
  LIBFOMUS_EXPORT FOMUS fomus_copy(FOMUS f);
  LIBFOMUS_EXPORT void fomus_save(FOMUS f, const char* filename); // INVALIDATES FOMUS INSTANCE, EVEN IF THERE's AN ERROR!
  
  // either/both of these can be NULL, newline means include newline
  LIBFOMUS_EXPORT void fomus_set_outputs(fomus_output out, fomus_output err, int newline);

#endif
  
#ifdef __cplusplus
}
#endif

#endif
