// -*- c++ -*-

/*
    Copyright (C) 2009  David Psenicka
    This file is part of FOMUS.

    FOMUS is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FOMUS is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef FOMUS_MODAPI_H
#define FOMUS_MODAPI_H

// --------------------------------------------------------------------------------
// C API for FOMUS modules

// Modules should include this and the API header file for whatever category it
// falls into (note spelling, voice distributing, etc.)

// FOMUS uses the FOMUS_MODULE_PATH environment variable to find all available
// modules
// --------------------------------------------------------------------------------

#include "basetypes.h"
#include "modtypes.h"

#ifdef __cplusplus
extern "C" {
#endif

  // --------------------------------------------------------------------------------
  // calls to FOMUS

  // module is responsible for creating/destroying all of its own strings
  // module is also responsible for handling rationals correctly (denominators must not equal 0!)
  
/*   struct module_setting* module_new_setting(); */
/*   // this frees any arrays inside the settings as well */
/*   void module_free_setting(struct module_setting* set); */

  LIBFOMUS_EXPORT struct module_list module_new_list(int n); // it's actually a vector, but in the config files it's a list
  // this frees any nested arrays inside the settings as well (very rarely need this)
  LIBFOMUS_EXPORT void module_free_list(struct module_list list);
  
  // SET functions--module can use these to set a module_value struct
  // automatically frees any lists (calls `module_free_list') if needed
  //void module_setval_bool(struct module_value* val, fomus_bool b);
  LIBFOMUS_EXPORT void module_setval_int(struct module_value* val, fomus_int i);
  LIBFOMUS_EXPORT void module_setval_float(struct module_value* val, fomus_float f);
  LIBFOMUS_EXPORT void module_setval_rat(struct module_value* val, struct fomus_rat r);
  LIBFOMUS_EXPORT void module_setval_string(struct module_value* val, const char* s); // module must keep track of (new/delete) its own strings!
  LIBFOMUS_EXPORT void module_setval_list(struct module_value* val, int n); // 
  // convience macros
  //#define SET_B module_setval_bool
#define SET_I module_setval_int
#define SET_F module_setval_float
#define SET_R module_setval_rat
#define SET_S module_setval_string
#define SET_L module_setval_list
  
  // UTILITY CONVERSION FUNCTIONS--these will return NULL or 0/0 or 0-length NULL list if conversion isn't possible (ie. converting from string or list)
  //fomus_bool module_getval_bool(const struct module_value* val);
  LIBFOMUS_EXPORT fomus_int module_getval_int(struct module_value val);
  LIBFOMUS_EXPORT fomus_float module_getval_float(struct module_value val);
  LIBFOMUS_EXPORT struct fomus_rat module_getval_rat(struct module_value val);
  LIBFOMUS_EXPORT const char* module_getval_string(struct module_value val);
  LIBFOMUS_EXPORT struct module_list module_getval_list(struct module_value val);
  // convience macros
  //#define GET_B module_getval_bool
#define GET_I module_getval_int
#define GET_F module_getval_float
#define GET_R module_getval_rat
#define GET_S module_getval_string
#define GET_L module_getval_list

  LIBFOMUS_EXPORT const char* module_valuetostr(struct module_value x);
  
  // rational utility functions
  inline struct fomus_rat fomus_makerat(fomus_int n, fomus_int d) {
    struct fomus_rat r = {n, d};
    return r;
  }
  LIBFOMUS_EXPORT void fomus_rat_reduce(struct fomus_rat* x);
  inline struct fomus_rat fomus_makerat_reduce(fomus_int n, fomus_int d) {
    struct fomus_rat r = {n, d};
    fomus_rat_reduce(&r);
    return r;
  }
  LIBFOMUS_EXPORT fomus_int fomus_rattoint(struct fomus_rat x);
  LIBFOMUS_EXPORT fomus_float fomus_rattofloat(struct fomus_rat x);
  LIBFOMUS_EXPORT struct fomus_rat fomus_inttorat(fomus_int x);
  LIBFOMUS_EXPORT struct fomus_rat fomus_floattorat(fomus_int x);
  LIBFOMUS_EXPORT int fomus_rat_eq(struct fomus_rat a, struct fomus_rat b);
  LIBFOMUS_EXPORT int fomus_rat_neq(struct fomus_rat a, struct fomus_rat b);
  LIBFOMUS_EXPORT int fomus_rat_lt(struct fomus_rat a, struct fomus_rat b);
  LIBFOMUS_EXPORT int fomus_rat_lteq(struct fomus_rat a, struct fomus_rat b);
  LIBFOMUS_EXPORT int fomus_rat_gt(struct fomus_rat a, struct fomus_rat b);
  LIBFOMUS_EXPORT int fomus_rat_gteq(struct fomus_rat a, struct fomus_rat b);
  LIBFOMUS_EXPORT struct fomus_rat fomus_rat_plus(struct fomus_rat a, struct fomus_rat b);
  LIBFOMUS_EXPORT struct fomus_rat fomus_rat_minus(struct fomus_rat a, struct fomus_rat b);
  LIBFOMUS_EXPORT struct fomus_rat fomus_rat_mult(struct fomus_rat a, struct fomus_rat b);
  LIBFOMUS_EXPORT struct fomus_rat fomus_rat_div(struct fomus_rat a, struct fomus_rat b);
  LIBFOMUS_EXPORT struct fomus_rat fomus_rat_inv(struct fomus_rat x);
  LIBFOMUS_EXPORT struct fomus_rat fomus_rat_neg(struct fomus_rat x);
  
  LIBFOMUS_EXPORT const char* module_rattostr(struct fomus_rat x);
  LIBFOMUS_EXPORT void module_free_string(const char* str);

  // utilities for validating numbers (for setting values)
  //int module_valid_bool(struct module_value val, valid_bool fun, const char* printerr); // fun is a callback function, called w/ each list value for add. checking
  LIBFOMUS_EXPORT int module_valid_int(struct module_value val, fomus_int min, enum module_bound minbnd, fomus_int max, enum module_bound maxbnd, valid_int fun, const char* printerr);
  LIBFOMUS_EXPORT int module_valid_rat(struct module_value val, struct fomus_rat min, enum module_bound minbnd, struct fomus_rat max, enum module_bound maxbnd, valid_rat fun, const char* printerr);
  LIBFOMUS_EXPORT int module_valid_num(struct module_value val, struct module_value min, enum module_bound minbnd, struct module_value max, enum module_bound maxbnd, valid_num fun, const char* printerr);
  LIBFOMUS_EXPORT int module_valid_string(struct module_value val, int minlen, int maxlen, valid_string fun, const char* printerr);
  
  // utilities for lists and strings--return 1 if invalid (not a list & doesn't contain all of specified type)
  //int module_valid_listofbools(struct module_value val, valid_listbool fun, const char* printerr); // fun is a callback function, called w/ each list value for add. checking
  LIBFOMUS_EXPORT int module_valid_listofints(struct module_value val, fomus_int min, enum module_bound minbnd, fomus_int max, enum module_bound maxbnd, valid_listint fun, const char* printerr);
  LIBFOMUS_EXPORT int module_valid_listofrats(struct module_value val, struct fomus_rat min, enum module_bound minbnd, struct fomus_rat max, enum module_bound maxbnd, valid_listrat fun, const char* printerr);
  LIBFOMUS_EXPORT int module_valid_listofnums(struct module_value val, struct module_value min, enum module_bound minbnd, struct module_value max, enum module_bound maxbnd, valid_listnum fun, const char* printerr);
  LIBFOMUS_EXPORT int module_valid_listofstrings(struct module_value val, int minlen, int maxlen, valid_liststring fun, const char* printerr);
  LIBFOMUS_EXPORT int module_valid_listofvals(struct module_value val, valid_listval fun, const char* printerr); // list of anything (or list of lists)
  
  // tests for list of unique keyword + value pairs of correct value type
  //int module_valid_maptobools(struct module_value val, valid_mapbool fun, const char* printerr); // fun is a callback function, called w/ each list value for add. checking
  LIBFOMUS_EXPORT int module_valid_maptoints(struct module_value val, fomus_int min, enum module_bound minbnd, fomus_int max, enum module_bound maxbnd, valid_mapint fun, const char* printerr);
  LIBFOMUS_EXPORT int module_valid_maptorats(struct module_value val, struct fomus_rat min, enum module_bound minbnd, struct fomus_rat max, enum module_bound maxbnd, valid_maprat fun, const char* printerr);
  LIBFOMUS_EXPORT int module_valid_maptonums(struct module_value val, struct module_value min, enum module_bound minbnd, struct module_value max, enum module_bound maxbnd, valid_mapnum fun, const char* printerr);
  LIBFOMUS_EXPORT int module_valid_maptostrings(struct module_value val, int minlen, int maxlen, valid_mapstring fun, const char* printerr);
  LIBFOMUS_EXPORT int module_valid_maptovals(struct module_value val, valid_mapval fun, const char* printerr);
  
  //void module_valid_err(const char* errtype); // just print an error (errtype = printerr string in above functions)
  
  // other utility functions
  LIBFOMUS_EXPORT int module_setting_allowed(const enum module_setting_loc allowedin, const enum module_setting_loc setloc);
  
  // GET SETTINGS for objects (notes, insts, parts, etc..)
  // make o = f for global variables
//   fomus_float module_get_fval(module_obj o, const char* set);
//   struct fomus_rat module_get_rval(module_obj o, const char* set);
//   fomus_int module_get_ival(module_obj o, const char* set);
//   const char* module_get_sval(module_obj o, const char* set);
//   struct module_value module_get_val(module_obj o, const char* set);

  LIBFOMUS_EXPORT int module_settingid(const char* set); // if getting setting value multiple times, store the integer ID for faster lookup

  // if o = 0, gets the default configuration file value
  LIBFOMUS_EXPORT fomus_float module_setting_fval(module_obj o, int id);
  LIBFOMUS_EXPORT struct fomus_rat module_setting_rval(module_obj o, int id);
  LIBFOMUS_EXPORT fomus_int module_setting_ival(module_obj o, int id);
  LIBFOMUS_EXPORT const char* module_setting_sval(module_obj o, int id);
  LIBFOMUS_EXPORT struct module_value module_setting_val(module_obj o, int id);

  LIBFOMUS_EXPORT fomus_float module_setting_minfval(module_obj o, int id);
  LIBFOMUS_EXPORT struct fomus_rat module_setting_minrval(module_obj o, int id);
  LIBFOMUS_EXPORT fomus_int module_setting_minival(module_obj o, int id);
  LIBFOMUS_EXPORT fomus_float module_setting_maxfval(module_obj o, int id);
  LIBFOMUS_EXPORT struct fomus_rat module_setting_maxrval(module_obj o, int id);
  LIBFOMUS_EXPORT fomus_int module_setting_maxival(module_obj o, int id);

  // GET LISTS OF OBJECTS!!!!!!  these aren't reentrant
  LIBFOMUS_EXPORT struct module_objlist module_get_percinsts(FOMUS f, int all);
  LIBFOMUS_EXPORT struct module_objlist module_get_insts(FOMUS f, int all);
  LIBFOMUS_EXPORT struct module_objlist module_get_parts(FOMUS f);  
  //LIBFOMUS_EXPORT struct module_objlist module_get_partsrefs(FOMUS f);

  // STDIN and STDOUT--use these so that output can be rerouted by program
  LIBFOMUS_EXPORT void module_stdout(const char* str, unsigned long n); // if n = 0, then str must be zero terminated
  LIBFOMUS_EXPORT void module_stderr(const char* str, unsigned long n); // if n = 0, then str must be zero terminated

#ifdef __cplusplus
}
#endif

// --------------------------------------------------------------------------------
// C++ stuff
#ifdef __cplusplus

inline fomus_rat operator-(const struct fomus_rat& x) {return fomus_rat_neg(x);}

inline bool operator==(const struct fomus_rat& a, const struct fomus_rat& b) {return fomus_rat_eq(a, b);}
inline bool operator!=(const struct fomus_rat& a, const struct fomus_rat& b) {return fomus_rat_neq(a, b);}
inline bool operator<(const struct fomus_rat& a, const struct fomus_rat& b) {return fomus_rat_lt(a, b);}
inline bool operator<=(const struct fomus_rat& a, const struct fomus_rat& b) {return fomus_rat_lteq(a, b);}
inline bool operator>(const struct fomus_rat& a, const struct fomus_rat& b) {return fomus_rat_gt(a, b);}
inline bool operator>=(const struct fomus_rat& a, const struct fomus_rat& b) {return fomus_rat_gteq(a, b);}
inline struct fomus_rat operator+(const struct fomus_rat& a, const struct fomus_rat& b) {return fomus_rat_plus(a, b);}
inline struct fomus_rat operator-(const struct fomus_rat& a, const struct fomus_rat& b) {return fomus_rat_minus(a, b);}
inline struct fomus_rat operator*(const struct fomus_rat& a, const struct fomus_rat& b) {return fomus_rat_mult(a, b);}
inline struct fomus_rat operator/(const struct fomus_rat& a, const struct fomus_rat& b) {return fomus_rat_div(a, b);}

inline bool operator==(const struct fomus_rat& a, const fomus_int b) {return fomus_rat_eq(a, fomus_inttorat(b));}
inline bool operator!=(const struct fomus_rat& a, const fomus_int b) {return fomus_rat_neq(a, fomus_inttorat(b));}
inline bool operator<(const struct fomus_rat& a, const fomus_int b) {return fomus_rat_lt(a, fomus_inttorat(b));}
inline bool operator<=(const struct fomus_rat& a, const fomus_int b) {return fomus_rat_lteq(a, fomus_inttorat(b));}
inline bool operator>(const struct fomus_rat& a, const fomus_int b) {return fomus_rat_gt(a, fomus_inttorat(b));}
inline bool operator>=(const struct fomus_rat& a, const fomus_int b) {return fomus_rat_gteq(a, fomus_inttorat(b));}
inline struct fomus_rat operator+(const struct fomus_rat& a, const fomus_int b) {return fomus_rat_plus(a, fomus_inttorat(b));}
inline struct fomus_rat operator-(const struct fomus_rat& a, const fomus_int b) {return fomus_rat_minus(a, fomus_inttorat(b));}
inline struct fomus_rat operator*(const struct fomus_rat& a, const fomus_int b) {return fomus_rat_mult(a, fomus_inttorat(b));}
inline struct fomus_rat operator/(const struct fomus_rat& a, const fomus_int b) {return fomus_rat_div(a, fomus_inttorat(b));}

inline bool operator==(const fomus_int a, const struct fomus_rat& b) {return fomus_rat_eq(fomus_inttorat(a), b);}
inline bool operator!=(const fomus_int a, const struct fomus_rat& b) {return fomus_rat_neq(fomus_inttorat(a), b);}
inline bool operator<(const fomus_int a, const struct fomus_rat& b) {return fomus_rat_lt(fomus_inttorat(a), b);}
inline bool operator<=(const fomus_int a, const struct fomus_rat& b) {return fomus_rat_lteq(fomus_inttorat(a), b);}
inline bool operator>(const fomus_int a, const struct fomus_rat& b) {return fomus_rat_gt(fomus_inttorat(a), b);}
inline bool operator>=(const fomus_int a, const struct fomus_rat& b) {return fomus_rat_gteq(fomus_inttorat(a), b);}
inline struct fomus_rat operator+(const fomus_int a, const struct fomus_rat& b) {return fomus_rat_plus(fomus_inttorat(a), b);}
inline struct fomus_rat operator-(const fomus_int a, const struct fomus_rat& b) {return fomus_rat_minus(fomus_inttorat(a), b);}
inline struct fomus_rat operator*(const fomus_int a, const struct fomus_rat& b) {return fomus_rat_mult(fomus_inttorat(a), b);}
inline struct fomus_rat operator/(const fomus_int a, const struct fomus_rat& b) {return fomus_rat_div(fomus_inttorat(a), b);}

inline bool operator==(const struct fomus_rat& a, const fomus_float b) {return fomus_rattofloat(a) == b;}
inline bool operator!=(const struct fomus_rat& a, const fomus_float b) {return fomus_rattofloat(a) != b;}
inline bool operator<(const struct fomus_rat& a, const fomus_float b) {return fomus_rattofloat(a) < b;}
inline bool operator<=(const struct fomus_rat& a, const fomus_float b) {return fomus_rattofloat(a) <= b;}
inline bool operator>(const struct fomus_rat& a, const fomus_float b) {return fomus_rattofloat(a) > b;}
inline bool operator>=(const struct fomus_rat& a, const fomus_float b) {return fomus_rattofloat(a) >= b;}
inline fomus_float operator+(const struct fomus_rat& a, const fomus_float b) {return fomus_rattofloat(a) + b;}
inline fomus_float operator-(const struct fomus_rat& a, const fomus_float b) {return fomus_rattofloat(a) - b;}
inline fomus_float operator*(const struct fomus_rat& a, const fomus_float b) {return fomus_rattofloat(a) * b;}
inline fomus_float operator/(const struct fomus_rat& a, const fomus_float b) {return fomus_rattofloat(a) / b;}

inline bool operator==(const fomus_float a, const struct fomus_rat& b) {return a == fomus_rattofloat(b);}
inline bool operator!=(const fomus_float a, const struct fomus_rat& b) {return a != fomus_rattofloat(b);}
inline bool operator<(const fomus_float a, const struct fomus_rat& b) {return a < fomus_rattofloat(b);}
inline bool operator<=(const fomus_float a, const struct fomus_rat& b) {return a <= fomus_rattofloat(b);}
inline bool operator>(const fomus_float a, const struct fomus_rat& b) {return a > fomus_rattofloat(b);}
inline bool operator>=(const fomus_float a, const struct fomus_rat& b) {return a >= fomus_rattofloat(b);}
inline fomus_float operator+(const fomus_float a, const struct fomus_rat& b) {return a + fomus_rattofloat(b);}
inline fomus_float operator-(const fomus_float a, const struct fomus_rat& b) {return a - fomus_rattofloat(b);}
inline fomus_float operator*(const fomus_float a, const struct fomus_rat& b) {return a * fomus_rattofloat(b);}
inline fomus_float operator/(const fomus_float a, const struct fomus_rat& b) {return a / fomus_rattofloat(b);}

struct fomus_exception {};
inline const module_value& fomus_make_val(const module_value& r) {return r;}
inline module_value fomus_make_val(const fomus_int x) {module_value r; r.type = module_int; r.val.i = x; return r;}
inline module_value fomus_make_val(const fomus_rat& x) {module_value r; r.type = module_rat; r.val.r = x; return r;}
inline module_value fomus_make_val(const fomus_float x) {module_value r; r.type = module_float; r.val.f = x; return r;}
inline module_value fomus_make_nullval() {module_value r; r.type = module_none; return r;}

inline module_value operator-(const module_value& a) {
  switch (a.type) {
  case module_int: return fomus_make_val(-a.val.i);
  case module_rat: return fomus_make_val(-a.val.r);
  case module_float: return fomus_make_val(-a.val.f);
  default: throw fomus_exception();
  }
}

template<typename T> inline bool operator==(const module_value& a, const T& b) {
  switch (a.type) {
  case module_int: return a.val.i == b;
  case module_rat: return a.val.r == b;
  case module_float: return a.val.f == b;
  default: throw fomus_exception();
  }
}
template<typename T> inline bool operator!=(const module_value& a, const T& b) {
  switch (a.type) {
  case module_int: return a.val.i != b;
  case module_rat: return a.val.r != b;
  case module_float: return a.val.f != b;
  default: throw fomus_exception();
  }
}
template<typename T> inline bool operator<(const module_value& a, const T& b) {
  switch (a.type) {
  case module_int: return a.val.i < b;
  case module_rat: return a.val.r < b;
  case module_float: return a.val.f < b;
  default: throw fomus_exception();
  }
}
template<typename T> inline bool operator<=(const module_value& a, const T& b) {
  switch (a.type) {
  case module_int: return a.val.i <= b;
  case module_rat: return a.val.r <= b;
  case module_float: return a.val.f <= b;
  default: throw fomus_exception();
  }
}
template<typename T> inline bool operator>(const module_value& a, const T& b) {
  switch (a.type) {
  case module_int: return a.val.i > b;
  case module_rat: return a.val.r > b;
  case module_float: return a.val.f > b;
  default: throw fomus_exception();
  }
}
template<typename T> inline bool operator>=(const module_value& a, const T& b) {
  switch (a.type) {
  case module_int: return a.val.i >= b;
  case module_rat: return a.val.r >= b;
  case module_float: return a.val.f >= b;
  default: throw fomus_exception();
  }
}
template<typename T> inline module_value operator+(const module_value& a, const T& b) {
  switch (a.type) {
  case module_int: return fomus_make_val(a.val.i + b);
  case module_rat: return fomus_make_val(a.val.r + b);
  case module_float: return fomus_make_val(a.val.f + b);
  default: throw fomus_exception();
  }
}
template<typename T> inline module_value operator-(const module_value& a, const T& b) {
  switch (a.type) {
  case module_int: return fomus_make_val(a.val.i - b);
  case module_rat: return fomus_make_val(a.val.r - b);
  case module_float: return fomus_make_val(a.val.f - b);
  default: throw fomus_exception();
  }
}
template<typename T> inline module_value operator*(const module_value& a, const T& b) {
  switch (a.type) {
  case module_int: return fomus_make_val(a.val.i * b);
  case module_rat: return fomus_make_val(a.val.r * b);
  case module_float: return fomus_make_val(a.val.f * b);
  default: throw fomus_exception();
  }
}
template<typename T> inline module_value operator/(const module_value& a, const T& b) {
  switch (a.type) {
  case module_int: return fomus_make_val(a.val.i / b);
  case module_rat: return fomus_make_val(a.val.r / b);
  case module_float: return fomus_make_val(a.val.f / b);
  default: throw fomus_exception();
  }
}
template<>
inline module_value operator/<fomus_int>(const module_value& a, const fomus_int& b) {
  switch (a.type) {
  case module_int: return (a.val.i % b) ? fomus_make_val(fomus_makerat_reduce(a.val.i, b)) : fomus_make_val(a.val.i / b);
  case module_rat: return fomus_make_val(a.val.r / b);
  case module_float: return fomus_make_val(a.val.f / b);
  default: throw fomus_exception();
  }
}

template<typename T> inline bool operator==(const T& a, const module_value& b) {
  switch (b.type) {
  case module_int: return a == b.val.i;
  case module_rat: return a == b.val.r;
  case module_float: return a == b.val.f;
  default: throw fomus_exception();
  }
}
template<typename T> inline bool operator!=(const T& a, const module_value& b) {
  switch (b.type) {
  case module_int: return a != b.val.i;
  case module_rat: return a != b.val.r;
  case module_float: return a != b.val.f;
  default: throw fomus_exception();
  }
}
template<typename T> inline bool operator<(const T& a, const module_value& b) {
  switch (b.type) {
  case module_int: return a < b.val.i;
  case module_rat: return a < b.val.r;
  case module_float: return a < b.val.f;
  default: throw fomus_exception();
  }
}
template<typename T> inline bool operator<=(const T& a, const module_value& b) {
  switch (b.type) {
  case module_int: return a <= b.val.i;
  case module_rat: return a <= b.val.r;
  case module_float: return a <= b.val.f;
  default: throw fomus_exception();
  }
}
template<typename T> inline bool operator>(const T& a, const module_value& b) {
  switch (b.type) {
  case module_int: return a > b.val.i;
  case module_rat: return a > b.val.r;
  case module_float: return a > b.val.f;
  default: throw fomus_exception();
  }
}
template<typename T> inline bool operator>=(const T& a, const module_value& b) {
  switch (b.type) {
  case module_int: return a >= b.val.i;
  case module_rat: return a >= b.val.r;
  case module_float: return a >= b.val.f;
  default: throw fomus_exception();
  }
}
template<typename T> inline module_value operator+(const T& a, const module_value& b) {
  switch (b.type) {
  case module_int: return fomus_make_val(a + b.val.i);
  case module_rat: return fomus_make_val(a + b.val.r);
  case module_float: return fomus_make_val(a + b.val.f);
  default: throw fomus_exception();
  }
}
template<typename T> inline module_value operator-(const T& a, const module_value& b) {
  switch (b.type) {
  case module_int: return fomus_make_val(a - b.val.i);
  case module_rat: return fomus_make_val(a - b.val.r);
  case module_float: return fomus_make_val(a - b.val.f);
  default: throw fomus_exception();
  }
}
template<typename T> inline module_value operator*(const T& a, const module_value& b) {
  switch (b.type) {
  case module_int: return fomus_make_val(a * b.val.i);
  case module_rat: return fomus_make_val(a * b.val.r);
  case module_float: return fomus_make_val(a * b.val.f);
  default: throw fomus_exception();
  }
}
template<typename T> inline module_value operator/(const T& a, const module_value& b) {
  switch (b.type) {
  case module_int: return fomus_make_val(a / b.val.i);
  case module_rat: return fomus_make_val(a / b.val.r);
  case module_float: return fomus_make_val(a / b.val.f);
  default: throw fomus_exception();
  }
}
template<>
inline module_value operator/<fomus_int>(const fomus_int& a, const module_value& b) {
  switch (b.type) {
  case module_int: return (a % b.val.i) ? fomus_make_val(fomus_makerat_reduce(a, b.val.i)) : fomus_make_val(a / b.val.i);
  case module_rat: return fomus_make_val(a / b.val.r);
  case module_float: return fomus_make_val(a / b.val.f);
  default: throw fomus_exception();
  }
}

inline bool operator==(const module_value& a, const module_value& b) {
  switch (b.type) {
  case module_int: return a == b.val.i;
  case module_rat: return a == b.val.r;
  case module_float: return a == b.val.f;
  default: throw fomus_exception();
  }
}
inline bool operator!=(const module_value& a, const module_value& b) {
  switch (b.type) {
  case module_int: return a != b.val.i;
  case module_rat: return a != b.val.r;
  case module_float: return a != b.val.f;
  default: throw fomus_exception();
  }
}
inline bool operator<(const module_value& a, const module_value& b) {
  switch (b.type) {
  case module_int: return a < b.val.i;
  case module_rat: return a < b.val.r;
  case module_float: return a < b.val.f;
  default: throw fomus_exception();
  }
}
inline bool operator<=(const module_value& a, const module_value& b) {
  switch (b.type) {
  case module_int: return a <= b.val.i;
  case module_rat: return a <= b.val.r;
  case module_float: return a <= b.val.f;
  default: throw fomus_exception();
  }
}
inline bool operator>(const module_value& a, const module_value& b) {
  switch (b.type) {
  case module_int: return a > b.val.i;
  case module_rat: return a > b.val.r;
  case module_float: return a > b.val.f;
  default: throw fomus_exception();
  }
}
inline bool operator>=(const module_value& a, const module_value& b) {
  switch (b.type) {
  case module_int: return a >= b.val.i;
  case module_rat: return a >= b.val.r;
  case module_float: return a >= b.val.f;
  default: throw fomus_exception();
  }
}
inline module_value operator+(const module_value& a, const module_value& b) {
  switch (b.type) {
  case module_int: return a + b.val.i;
  case module_rat: return a + b.val.r;
  case module_float: return a + b.val.f;
  default: throw fomus_exception();
  }
}
inline module_value operator-(const module_value& a, const module_value& b) {
  switch (b.type) {
  case module_int: return a - b.val.i;
  case module_rat: return a - b.val.r;
  case module_float: return a - b.val.f;
  default: throw fomus_exception();
  }
}
inline module_value operator*(const module_value& a, const module_value& b) {
  switch (b.type) {
  case module_int: return a * b.val.i;
  case module_rat: return a * b.val.r;
  case module_float: return a * b.val.f;
  default: throw fomus_exception();
  }
}
inline module_value operator/(const module_value& a, const module_value& b) {
  switch (b.type) {
  case module_int: return a / b.val.i;
  case module_rat: return a / b.val.r;
  case module_float: return a / b.val.f;
  default: throw fomus_exception();
  }
}

#endif

#endif
