// -*- c++ -*-

/*
    Copyright (C) 2009  David Psenicka
    This file is part of FOMUS.

    FOMUS is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FOMUS is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef FOMUS_MODCBACKS_H
#define FOMUS_MODCBACKS_H

// --------------------------------------------------------------------------------
// C API for FOMUS modules

// Modules should include this and the API header file for whatever category it
// falls into (note spelling, voice distributing, etc.)

// FOMUS uses the FOMUS_MODULE_PATH environment variable to find all available
// modules
// --------------------------------------------------------------------------------

#include "basetypes.h"
#include "modtypes.h"

#ifdef __cplusplus
extern "C" {
#endif
  
  // --------------------------------------------------------------------------------
  // callbacks from FOMUS
  // none of these functions should report errors throw module_err callback

#ifndef FOMUSMOD_HIDE

  // initialization and unloading
  FOMUSMOD_EXPORT void module_init();
  FOMUSMOD_EXPORT void module_free();

  // return a name for this module
  FOMUSMOD_EXPORT const char* module_longname();
  FOMUSMOD_EXPORT const char* module_author();
  // return a documentation string
  FOMUSMOD_EXPORT const char* module_doc();
  // report any errors with an error string or 0 if none
  // should also reset itself
  FOMUSMOD_EXPORT const char* module_initerr();
/*   void module_errreset(); */
  
  // return the module category
  FOMUSMOD_EXPORT enum module_type module_type();

  // send variables to FOMUS--once these are registered, module code must
  // call FOMUS to get the values since they might change depending on context
  // fomus free the entire array and arrays in the variable structs.  fomus won't
  // delete any character strings
  //int module_n_settings(); // return number of settings
  FOMUSMOD_EXPORT int module_get_setting(int n, struct module_setting* set, int id);
  // called after all settings are processed
  FOMUSMOD_EXPORT void module_ready();

#endif
  
//   /*enum module_iter_type*/ int module_get_itertype();

#ifdef __cplusplus
}
#endif

#endif
