// -*- c++ -*-

/*
    Copyright (C) 2009  David Psenicka
    This file is part of FOMUS.

    FOMUS is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FOMUS is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef FOMUS_MODENGINE_H
#define FOMUS_MODENGINE_H

// --------------------------------------------------------------------------------
// C API for FOMUS ENGINES

// An engine must provide it's own header file for other modules to use it
// function addresses must be provided through a structure passed to FOMUS which is
// then passed to the modules that want to use it.

// Engines that share the same interface should must the same header file
// --------------------------------------------------------------------------------

#include "modinit.h"
#include "moderr.h"

#ifdef __cplusplus
extern "C" {
#endif

#ifndef FOMUSMOD_HIDE

// initialization and unloading (different from module init, free--these are called
// once per score)
// void engine_init();
// void engine_free();

/* // this is the engine's own data instance--this is called before the new_interface function-- */
/* // it should create a structure that contains an interface structure as part of it */
/* void* engine_new_data(); */
/* void engine_free_data(void* data); */

// return a pointer to a structure containing all necessary functionality for the
// engine.  FOMUS calls this only once and passes it to the modules that want to use it.
// the modules should initialize it with callback functions and data so the engine
// can call the appropriate functions when stepped.
// the contents should be defined in the engine's header file
  FOMUSMOD_EXPORT void* engine_get_iface(void* engdata);
// void engine_ready(void* data); // called after interface info filled so engine can do further initializations

// different engines can have (and should have if possible) compatible interfaces.
// each engine header file (ie. unique interface) should have a macro defined to
// a unique integer value and this function should return it.  This insures that
// FOMUS's modules will know when they can use this engine in place of ones with 
// identical interfaces
  FOMUSMOD_EXPORT int engine_ifaceid();

// engine must handle and reference pointer values
/* typedef void* engine_node; */

// run the engine--engine returns when finished--must check other modules for errors and exit on error
  FOMUSMOD_EXPORT void engine_run(void* data);

#endif
  
#ifdef __cplusplus
}
#endif

#endif
