// -*- c++ -*-

/*
    Copyright (C) 2009  David Psenicka
    This file is part of FOMUS.

    FOMUS is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FOMUS is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef FOMUS_MODNOTES_H
#define FOMUS_MODNOTES_H

#include "modtypes.h" // module_obj
#include "modenums.h"

#ifdef __cplusplus
extern "C" {
#endif

  #define FOMUS_MODAPI_VERSION 0
    
  struct module_intslist {
    int n;
    const int* ints;
  };
  struct module_ratslist {
    int n;
    const struct fomus_rat* rats;
  };
  struct module_markslist {
    int n;
    const module_markobj* marks;
  };

  //LIBFOMUS_EXPORT int module_modapi_version();
  
  LIBFOMUS_EXPORT enum parts_grouptype module_partgroupbegin(module_partobj part, int lvl);
  LIBFOMUS_EXPORT int module_partgroupend(module_partobj part, int lvl);
  
  // get the next note, or NULL if at end
  LIBFOMUS_EXPORT module_noteobj module_nextnote();
  LIBFOMUS_EXPORT module_measobj module_nextmeas(); // for modules working with entire parts--this isn't synchronized with `nextnote', these are just iterators
  LIBFOMUS_EXPORT module_partobj module_nextpart();
  // can send NULL to these to get first one
  LIBFOMUS_EXPORT module_noteobj module_peeknextnote(module_noteobj note); // next measure or note--gets measures/notes out of sequence from above
  LIBFOMUS_EXPORT module_measobj module_peeknextmeas(module_measobj meas); // next measure or note--gets measures/notes out of sequence from above
  LIBFOMUS_EXPORT module_partobj module_peeknextpart(module_partobj part);

  LIBFOMUS_EXPORT module_measobj module_meas(module_noteobj note); // get parent object
  LIBFOMUS_EXPORT module_partobj module_part(module_obj obj); // get parent object
  
  LIBFOMUS_EXPORT struct module_value module_vtime(module_obj obj); // measure or note
  LIBFOMUS_EXPORT struct fomus_rat module_time(module_obj obj); // measure or note
  LIBFOMUS_EXPORT struct module_value module_vgracetime(module_obj obj); // measure or note--could return module_none (if not grace note!)
  LIBFOMUS_EXPORT struct fomus_rat module_gracetime(module_obj obj); // measure or note--could return module_none (if not grace note!)
  LIBFOMUS_EXPORT struct module_value module_vdur(module_obj obj); // grace notes have dur = 0 and endtime = time
  LIBFOMUS_EXPORT struct fomus_rat module_dur(module_obj obj); // grace notes have dur = 0 and endtime = time
  LIBFOMUS_EXPORT struct fomus_rat module_gracedur(module_obj obj);
  LIBFOMUS_EXPORT struct module_value module_vgracedur(module_noteobj note);
  LIBFOMUS_EXPORT struct module_value module_vendtime(module_obj obj); // --could return module_none (if not grace note!)
  LIBFOMUS_EXPORT struct fomus_rat module_endtime(module_obj obj); // --could return module_none (if not grace note!)
  LIBFOMUS_EXPORT struct module_value module_vgraceendtime(module_obj obj);
  LIBFOMUS_EXPORT struct fomus_rat module_graceendtime(module_obj obj);
  LIBFOMUS_EXPORT struct module_value module_vtiedendtime(module_noteobj note);
  LIBFOMUS_EXPORT struct fomus_rat module_tiedendtime(module_noteobj note);
  LIBFOMUS_EXPORT struct module_value module_vnote(module_noteobj note);
  LIBFOMUS_EXPORT struct fomus_rat module_note(module_noteobj note);
  LIBFOMUS_EXPORT struct module_value module_dyn(module_noteobj note); // grace notes have dur = 0 and endtime = time
  LIBFOMUS_EXPORT fomus_int module_writtennote(module_noteobj note); // with accidentals applied--should be a white note 
  LIBFOMUS_EXPORT int module_voice(module_noteobj note); // gets first one if there's still a list, otherwise return 0 (-1 is reserved for special cases)
  LIBFOMUS_EXPORT int module_staff(module_noteobj note); // gets first one if there's still a list, otherwise return 0
  LIBFOMUS_EXPORT struct module_intslist module_voices(module_obj obj); // for parts, measures or notes--returns SORTED list
  LIBFOMUS_EXPORT struct module_intslist module_staves(module_obj obj); // for parts, measures or notes--returns SORTED list of ONLY the staves that are occupied
  // possible clefs obtained by going down to all lists of clefs in staves in instr--if note, then take extra note clef setting into account
  LIBFOMUS_EXPORT struct module_intslist module_clefs(module_obj obj); // for parts, measures or notes--returns SORTED list
  LIBFOMUS_EXPORT struct module_intslist module_clefsinstaff(module_obj obj, int staff); // for parts, measures or notes--returns SORTED list
  LIBFOMUS_EXPORT int module_octsign(module_noteobj note); // -1 = octave down, 1 = octave up, etc..
  LIBFOMUS_EXPORT struct fomus_rat module_fullwrittenacc(module_noteobj note); // returns FOMUS_INT_MAX/1 if no accidental
  LIBFOMUS_EXPORT struct fomus_rat module_writtenacc1(module_noteobj note); // returns FOMUS_INT_MAX/1 if no accidental
  LIBFOMUS_EXPORT struct fomus_rat module_writtenacc2(module_noteobj note); // returns FOMUS_INT_MAX/1 if no accidental
  LIBFOMUS_EXPORT struct fomus_rat module_fullacc(module_noteobj note);
  LIBFOMUS_EXPORT struct fomus_rat module_acc1(module_noteobj note);
  LIBFOMUS_EXPORT struct fomus_rat module_acc2(module_noteobj note);
  LIBFOMUS_EXPORT struct fomus_rat module_adjdur(module_obj obj, int level); // level is tuplet level, 1/8 = eighth note, 1/16 = sixteenth, etc., if level = -1, all levels
  LIBFOMUS_EXPORT struct fomus_rat module_adjgracedur(module_obj obj, int level); // level is tuplet level, 1/8 = eighth note, 1/16 = sixteenth, etc.
  LIBFOMUS_EXPORT struct fomus_rat module_beatstoadjdur(module_obj obj, struct fomus_rat dur, int level);
  //fomus_rat module_beatstowrittengracedur(module_obj obj, fomus_rat dur, int level);

  //   #warning "combine these w/ chordlow, beginchord, etc..--fix usage in beams.cc"
  LIBFOMUS_EXPORT int module_octavebegin(module_noteobj obj); // notes in same chord + staffn have same octave sign
  LIBFOMUS_EXPORT int module_octaveend(module_noteobj obj);

  LIBFOMUS_EXPORT int module_istiedleft(module_noteobj note);
  LIBFOMUS_EXPORT int module_istiedright(module_noteobj note);

  LIBFOMUS_EXPORT module_noteobj module_leftmosttiednote(module_noteobj note);
  
  LIBFOMUS_EXPORT int module_tupletbegin(module_noteobj note, int level); // begin or end of tuplet?--should return same for ea. note in chord
  LIBFOMUS_EXPORT int module_tupletend(module_noteobj note, int level);
  LIBFOMUS_EXPORT struct fomus_rat module_tuplet(module_noteobj note, int level); // get the tuple--rational is NOT normalized--returns 0 if no tuplet at that level
  LIBFOMUS_EXPORT struct fomus_rat module_fulltupdur(module_noteobj note, int level); // get full duration span of tuplet at that level
  
  LIBFOMUS_EXPORT int module_hasvoice(module_noteobj note, int voice);
  LIBFOMUS_EXPORT int module_hasstaff(module_noteobj note, int staff); // (in possible choices)
  LIBFOMUS_EXPORT int module_hasclef(module_obj obj, int clef); // (in possible choices)
  LIBFOMUS_EXPORT int module_hasstaffclef(module_obj obj, int clef, int staff); // (in possible choices)
    
  LIBFOMUS_EXPORT int module_isrest(module_noteobj note);
  LIBFOMUS_EXPORT int module_isnote(module_noteobj note);
  LIBFOMUS_EXPORT int module_isgrace(module_noteobj note);
  LIBFOMUS_EXPORT int module_less(module_obj obj1, module_obj obj2);
  
  LIBFOMUS_EXPORT struct module_ratslist module_divs(module_measobj meas); // get the initial dividing points of the measure

  LIBFOMUS_EXPORT int module_clefmidpitch(int clef);

  LIBFOMUS_EXPORT struct module_markslist module_singlemarks(module_noteobj note);
  LIBFOMUS_EXPORT struct module_markslist module_spannerbegins(module_noteobj note);
  LIBFOMUS_EXPORT struct module_markslist module_spannerends(module_noteobj note);
  LIBFOMUS_EXPORT struct module_markslist module_marks(module_noteobj note);
  LIBFOMUS_EXPORT int module_markid(module_markobj mark); // return a mark id
  LIBFOMUS_EXPORT int module_markcantouch(int id);
  LIBFOMUS_EXPORT int module_markcanspanone(int id);
  LIBFOMUS_EXPORT int module_markisvoice(int id);
  LIBFOMUS_EXPORT int module_markisstaff(int id);
  LIBFOMUS_EXPORT int module_markcanreduce(int id);
  LIBFOMUS_EXPORT int module_markcanspanrests(int id);
  LIBFOMUS_EXPORT int module_markcanendonrests(int id);
  LIBFOMUS_EXPORT int module_markisconflictingprev(int id); // if true, this spanner is quivalent to the spanner w/ id - 2
  LIBFOMUS_EXPORT int module_markisconflictingnext(int id); // if true, this spanner is quivalent to the spanner w/ id + 2
  LIBFOMUS_EXPORT const char* module_markstring(module_markobj mark); // optional string argument
  LIBFOMUS_EXPORT struct module_value module_marknum(module_markobj mark); // optional number argument

  LIBFOMUS_EXPORT const char* module_pitchtostr(struct module_value pitch);
  LIBFOMUS_EXPORT const char* module_pitchnametostr(struct module_value pitch, const int forceacc);
  LIBFOMUS_EXPORT const char* module_id(module_obj obj); // name of any object (part, instr, etc..)
  
  LIBFOMUS_EXPORT int module_isbeginchord(module_noteobj note);
  LIBFOMUS_EXPORT int module_isendchord(module_noteobj note);
  LIBFOMUS_EXPORT int module_ischordlow(module_noteobj note);
  LIBFOMUS_EXPORT int module_ischordhigh(module_noteobj note);
  LIBFOMUS_EXPORT int module_isperc(module_noteobj note);
  
  LIBFOMUS_EXPORT void module_skipassign(module_noteobj note);

  //module_obj module_measdef(module_measobj meas); // get attributes

  // use an auxiliary module
  LIBFOMUS_EXPORT void module_get_auxiface(const char* modname, int ifaceid, void* iface);
  
  LIBFOMUS_EXPORT struct fomus_rat module_timesig(module_measobj meas); // the time signature (rat isn't reduced)
  LIBFOMUS_EXPORT struct fomus_rat module_writtenmult(module_measobj meas); // multiplier to get the ACTUAL written dur

  LIBFOMUS_EXPORT int module_clef(module_noteobj note);
  LIBFOMUS_EXPORT int module_totalnstaves(module_obj obj); // for instruments and parts--returns actual number of staves on score (not list of used staves!)
  
  LIBFOMUS_EXPORT int str_to_clef(const char* str);
  LIBFOMUS_EXPORT const char* clef_to_str(int en);
  LIBFOMUS_EXPORT int str_to_mark(const char* str);
  LIBFOMUS_EXPORT const char* mark_to_str(int id);
  
  LIBFOMUS_EXPORT const char* module_staffclef(module_partobj part, int staff); // gets the `clef' setting from the `1st defined clef in the staff in the given part (ie. the "default" clef)
  LIBFOMUS_EXPORT int module_staffvoice(module_noteobj note);
  
  LIBFOMUS_EXPORT int module_isfullmeasrest(module_measobj meas);
  
  LIBFOMUS_EXPORT const char* module_getposstring(module_noteobj note); // print the file position to the module stdout

  struct module_noteparts {
    struct fomus_rat note, acc1, acc2, oct;
  };
  LIBFOMUS_EXPORT struct fomus_rat module_str_to_note(const char* str, struct module_noteparts* parts); // might return FOMUS_INT_MAX/1!
  LIBFOMUS_EXPORT struct fomus_rat module_str_to_acc(const char* str, struct module_noteparts* parts); // might return FOMUS_INT_MAX/1!

  struct module_keysigref {
    struct fomus_rat acc1, acc2;
  };
  // always returns array of 75 (diatonic notes)
  LIBFOMUS_EXPORT const struct module_keysigref* module_keysigref(module_measobj meas);
  LIBFOMUS_EXPORT struct module_keysigref module_keysigacc(module_noteobj note);
  LIBFOMUS_EXPORT struct module_keysigref module_measkeysigacc(module_measobj meas, int note);
  
#ifdef __cplusplus
}
#endif  

#endif
