// -*- c++ -*-

/*
    Copyright (C) 2009  David Psenicka
    This file is part of FOMUS.

    FOMUS is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FOMUS is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef FOMUS_MODOUT_H
#define FOMUS_MODOUT_H

#include "basetypes.h"
#include "modinout.h"
#include "modnotes.h"
#include "modmod.h"

#ifdef __cplusplus
extern "C" {
#endif

  enum modout_sorttype {
    modout_part,
    modout_measure,
    modout_staff,
    modout_voice,
    modout_offset
  };
  struct modout_sorttypes {
    int n;
    enum modout_sorttype* types;
  };
  
#ifndef FOMUSMOD_HIDE
  
  // CALLBACKS
  // return lowercase filename extension strings (0 when the last is reached)
  FOMUSMOD_EXPORT const char* modout_get_extension(int n);
  // return a loadid that can be used instead of a filename extension (this is checked first!)--return 0 if none (=file input)
  FOMUSMOD_EXPORT const char* modout_get_saveid(); // for non-file types, return a string identifier (ex. midiin...) not implemented yet

  FOMUSMOD_EXPORT fomus_bool modout_ispre(); // whether or not module writes before processing... usually return false here (fmsout.cc returns true)

  FOMUSMOD_EXPORT struct modout_sorttypes modout_sorttype(); // how everything is to be sorted
  
  // data is data object returned by newdata function
  // filename is complete (and should be in OS's native format) and exists
  FOMUSMOD_EXPORT void modout_write(FOMUS f, void* data, const char* filename);

#endif
  
  // SPECIAL FUNCTIONS (TOO SPECIAL FOR MODNOTES.H)
  LIBFOMUS_EXPORT int modout_beamsleft(module_noteobj note); // returns number of bars on beam
  LIBFOMUS_EXPORT int modout_beamsright(module_noteobj note);
  LIBFOMUS_EXPORT int modout_connbeamsleft(module_noteobj note); // returns number of connected bars on beam
  LIBFOMUS_EXPORT int modout_connbeamsright(module_noteobj note);

  enum module_keysig_modetype {
    keysig_none,
    keysig_common_maj, // a circle-of-fifths maj/min key signature
    keysig_common_min, 
    keysig_indiv, // vector of up to 7 indivual notes in 1 octave
    keysig_fullindiv // vector of up to 75 notes (= diatonic equiv of 128 chromatic notes)
  };
  struct module_keysig_indiv {
    int dianote; // -1 if some user-caused error
    struct fomus_rat acc1, acc2;
  };
  struct module_keysig {
    enum module_keysig_modetype mode;
    int dianote, acc; // used only if common-the diatonic note and acc
    int n;
    const struct module_keysig_indiv* indiv; // always used (common still fills it for reference), filled in order defined
    //module_keysig_ref* ref; // always filled vector look-up table (75 diatonic notes)
  };
  LIBFOMUS_EXPORT struct module_keysig modout_keysigdef(module_measobj meas);
  LIBFOMUS_EXPORT int modout_keysigequal(struct module_keysig key1, struct module_keysig key2);

  LIBFOMUS_EXPORT int modout_markorder(module_markobj mark); // usually is 1--spanner end marks might be 0 (beginning of note) or 2 (end of note), begin marks are 1 (beginning of note, but after 0)
  
#ifdef __cplusplus
}
#endif

#endif
