// -*- c++ -*-

/*
    Copyright (C) 2009  David Psenicka
    This file is part of FOMUS.

    FOMUS is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FOMUS is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef FOMUS_MODTYPES_H
#define FOMUS_MODTYPES_H

// --------------------------------------------------------------------------------
// C API for FOMUS modules

// Modules should include this and the API header file for whatever category it
// falls into (note spelling, voice distributing, etc.)

// FOMUS uses the FOMUS_MODULE_PATH environment variable to find all available
// modules
// --------------------------------------------------------------------------------

#include "basetypes.h"

#ifdef __cplusplus
extern "C" {
#endif
  
  // --------------------------------------------------------------------------------
  // types

  typedef void* module_obj;

  typedef void* module_noteobj;
  typedef void* module_measobj;
  typedef void* module_partobj;
  typedef void* module_markobj;
  typedef void* module_instobj;
  typedef void* module_staffobj;
  typedef void* module_clefobj;
  
  // module categories
  // UPDATE USERSTRS.CC!!!!!!!!!  DON'T CHANGE THE ORDER
  enum module_type {
    module_nomodtype, // used only for info queries
    module_modinternal,
    module_modinput,
    module_modoutput,
    module_modengine,
    module_modaux,
    module_modmeas,
    module_modtquant,
    module_modpquant,
    module_modcheck,
    module_modtpose,
    module_modvoices,
    module_modmarkevs,
    module_modprune,
    module_modvmarks, 
    module_modaccs,
    module_modcautaccs,
    module_modstaves,
    module_modrstaves,
    module_modsmarks, 
    module_modocts,
    module_moddynamics, 
    module_moddivide,
    module_modmerge, 
    module_modbeams,
    module_modmarks,
    module_modmarklayouts, 
    module_modparts,
    module_modmetaparts,
    module_modpercnotes,
    module_modspecial 
  };

  // fomus will pass the fomus_data object to `run' function if module_all,
  // a part object if module_bypart, and a meas object if module_bymeas
  enum module_iter_type {
    module_all = 0x0, // these groups can be combined (ex. by part & by voice)
    module_bypart = 0x1,
    module_bymeas = 0x2,
    module_bymeasgroups = 0x4,

    module_byvoice = 0x8, // everything else is either/or or neither
    module_bystaff = 0x10,

    module_nograce = 0x20,
    module_graceonly = 0x40,
    
    module_norests = 0x80,
    module_restsonly = 0x100,

    module_firsttied = 0x200, // combine these three, or none for all
    module_midtied = 0x400,
    module_lasttied = 0x800,

    module_noperc = 0x1000,
    module_perconly = 0x2000
  };
  
  // types and structs for defining module variables
  enum module_value_type {
    module_none, // used only for info queries, not by modules!
    
    module_bool, // for SETTING TYPE only, same as an int, only parsing is different
    module_int,
    module_float,
    module_rat,
    module_number, // for SETTING TYPE only, NOT a VALUE TYPE
    module_notesym, // for SETTING TYPE only
    module_string,
    module_stringnum, // for MARKS only--args can include both a string and a number
    
    module_list, // everything after this must be some kind of list! (for `freevalue' function to catch possible errors)
    module_list_nums, // for setting, must use more specific list type...
    module_list_strings,
    //module_list_any,
    module_symmap_nums,
    module_symmap_strings,
    module_list_numlists,
    module_list_stringlists,
    module_symmap_numlists,
    module_symmap_stringlists,
    //module_symmap_any,
    //module_arglist
    module_special
  };
  struct module_value;
  struct module_list {
    int n;
    struct module_value* vals;
  };
  union module_value_val {
    //fomus_bool b; // bool
    fomus_int i; // integer
    fomus_float f; // float
    struct fomus_rat r; // rational
    const char* s; // string
    struct module_list l; // array
  };
  struct module_value {
    enum module_value_type type; // the type
    union module_value_val val;
  };
  
  // type for variable validation function
  typedef int (*module_valid_fun)(struct module_value val); // must return true/false and optionally set ERR string to error message
  //#warning "get rid of this, not doing deps"
  //typedef int (*module_validdeps_fun)(FOMUS f, struct module_value val); // must return true/false and optionally set ERR string to error message
  
  // allowed locations
  enum module_setting_loc {
    module_noloc, // used only for info queries, not by modules!

    module_locglobal,
    
    module_locpart, // implies global
    module_locinst,
    module_locpercinst,
    module_locpartmap,

    module_locmeasdef,

    module_locimport,
    module_locexport,

    module_locstaff,
    module_locclef,
    
    module_locnote, // implies instpartmeas and global
  };
  
  // struct for registering a FOMUS variable
  struct module_setting {
    const char* name; // unique variable name, mandatory
    enum module_value_type type;
    const char* descdoc; // documentation string (can be NULL)
    const char* typedoc; // can be NULL
    struct module_value val; // default value
    enum module_setting_loc loc; // bit ORed allowed locations
    module_valid_fun valid; // can be NULL--this is usually the one to set
    //module_validdeps_fun validdeps; // can be NULL, only set to check it in relation to dependencies
    int uselevel; // if other than 0 (often/always set), set to 1-3 (1=details like number of chars in tab space, 2=sometimes, 3=seldom used/expert)
  };

//   enum {
//     module_noprop = 0x0,

//     module_single = 0x0,
//     module_begin = 0x10,
//     module_end = 0x01
//   };
  struct module_mark {
    const char* name;
    enum module_value_type type; // none, numb, string or stringnumb
    unsigned int props;
    const char* doc;
  };

  // function types for various utility functions
  //typedef int (*valid_bool)(fomus_bool val);
  typedef int (*valid_int)(fomus_int val);
  typedef int (*valid_rat)(struct fomus_rat val);
  typedef int (*valid_num)(struct module_value val);
  typedef int (*valid_string)(const char* val);

  //typedef int (*valid_listbool)(int n, fomus_bool val);
  typedef int (*valid_listint)(int n, fomus_int val);
  typedef int (*valid_listrat)(int n, struct fomus_rat val);
  typedef int (*valid_listnum)(int n, const struct module_value* val);
  typedef int (*valid_liststring)(int n, const char* val);
  typedef int (*valid_listval)(int n, struct module_value val);

  //typedef int (*valid_mapbool)(int n, const char* sym, fomus_bool val);
  typedef int (*valid_mapint)(int n, const char* sym, fomus_int val);
  typedef int (*valid_maprat)(int n, const char* sym, struct fomus_rat val);
  typedef int (*valid_mapnum)(int n, const char* sym, const struct module_value* val);
  typedef int (*valid_mapstring)(int n, const char* sym, const char* val);
  typedef int (*valid_mapval)(int n, const char* sym, struct module_value val);

  enum module_bound {
    module_nobound,
    module_incl,
    module_excl
  };

  // for passing lists of objects
  struct module_objlist {
    int n;
    module_obj* objs;
  };

#ifdef __cplusplus
}
#endif

#endif
