;;   Copyright (C) 2009  David Psenicka
;;   This file is part of FOMUS.

;;   FOMUS is free software: you can redistribute it and/or modify
;;   it under the terms of the GNU General Public License as published by
;;   the Free Software Foundation, either version 3 of the License, or
;;   (at your option) any later version.

;;   FOMUS is distributed in the hope that it will be useful,
;;   but WITHOUT ANY WARRANTY; without even the implied warranty of
;;   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;   GNU General Public License for more details.

;;   You should have received a copy of the GNU General Public License
;;   along with this program.  If not, see <http://www.gnu.org/licenses/>.

;; ------------------------------------------------------------------------------------------------------------------------
;; FOMUS INTERFACE FOR LISP

(defpackage #:fomus
  (:nicknames #:fms)
  (:use #:common-lisp)
  (:export #:fomus-load #:fomus-run #:fomus-version
	   #:fomus-note #:fomus-rest #:fomus-mark #:fomus-meas
	   #:fomus-measattr #:fomus-part #:fomus-metapart #:fomus-partsref #:fomus-inst #:fomus-percinst
	   #:fomus-setting
	   #:with-fomus-score
	   ))

;; #+fomus-usepackage
;; (eval-when (:compile-toplevel :load-toplevel :execute)
;;   (use-package 'fomus 'common-lisp-user))
(provide :fomus)
(in-package #:fomus)

;; LOAD LIBRARY
#+cffi
(eval-when (:compile-toplevel :load-toplevel :execute)
  (defvar *so-notloaded* t)
  (when *so-notloaded*
    (cffi:define-foreign-library libfomus
	(t (:default "libfomus")))
    (cffi:use-foreign-library libfomus)
    (setf *so-notloaded* nil)))

;; ------------------------------------------------------------------------------------------------------------------------


(cffi:defcstruct fomus_rat
	(num :long)
	(den :long))

(cl:defconstant FOMUS_API_VERSION 0)

(cffi:defcenum fomus_param
	:fomus_par_none
	:fomus_par_entry
	:fomus_par_list
	:fomus_par_setting
	:fomus_par_settingval
	:fomus_par_locfile
	:fomus_par_locline
	:fomus_par_loccol
	:fomus_par_import
	:fomus_par_import_settingval
	:fomus_par_export
	:fomus_par_export_settingval
	:fomus_par_clef
	:fomus_par_clef_settingval
	:fomus_par_staff
	:fomus_par_staff_clefs
	:fomus_par_staff_settingval
	:fomus_par_percinst
	:fomus_par_percinst_template
	:fomus_par_percinst_id
	:fomus_par_percinst_imports
	:fomus_par_percinst_export
	:fomus_par_percinst_settingval
	:fomus_par_inst
	:fomus_par_inst_template
	:fomus_par_inst_id
	:fomus_par_inst_staves
	:fomus_par_inst_imports
	:fomus_par_inst_export
	:fomus_par_inst_percinsts
	:fomus_par_inst_settingval
	:fomus_par_part
	:fomus_par_part_id
	:fomus_par_part_inst
	:fomus_par_part_settingval
	:fomus_par_partmap
	:fomus_par_partmap_part
	:fomus_par_partmap_metapart
	:fomus_par_partmap_settingval
	:fomus_par_metapart
	:fomus_par_metapart_id
	:fomus_par_metapart_partmaps
	:fomus_par_metapart_settingval
	:fomus_par_measdef
	:fomus_par_measdef_id
	:fomus_par_measdef_settingval
	:fomus_par_list_percinsts
	:fomus_par_list_insts
	:fomus_par_settingval_percinsts
	:fomus_par_settingval_insts
	:fomus_par_region
	:fomus_par_region_voice
	:fomus_par_region_voicelist
	:fomus_par_region_time
	:fomus_par_region_gracetime
	:fomus_par_region_duration
	:fomus_par_region_pitch
	:fomus_par_region_dynlevel
	:fomus_par_region_mark
	:fomus_par_region_settingval
	:fomus_par_time
	:fomus_par_gracetime
	:fomus_par_duration
	:fomus_par_pitch
	:fomus_par_dynlevel
	:fomus_par_voice
	:fomus_par_note_settingval
	:fomus_par_meas
	:fomus_par_meas_measdef
	:fomus_par_markid
	:fomus_par_markval
	:fomus_par_mark
	:fomus_par_noteevent
	:fomus_par_restevent
	:fomus_par_markevent
	:fomus_par_events
	:fomus_par_n)

(cffi:defcenum fomus_action
	:fomus_act_none
	:fomus_act_set
	:fomus_act_inc
	:fomus_act_dec
	:fomus_act_mult
	:fomus_act_div
	:fomus_act_start
	:fomus_act_append
	:fomus_act_add
	:fomus_act_remove
	:fomus_act_end
	:fomus_act_clear
	:fomus_act_queue
	:fomus_act_cancel
	:fomus_act_resume
	:fomus_act_n)

(cffi:defcfun ("fomus_api_version" fomus_api_version) :int)

(cffi:defcvar ("fomus_err" fomus_err)
 :int)

(cffi:defcfun ("fomus_version" fomus_version) :string)

(cffi:defcfun ("fomus_init" fomus_init) :void)

(cffi:defcfun ("fomus_new" fomus_new) :pointer)

(cffi:defcfun ("fomus_free" fomus_free) :void
  (f :pointer))

(cffi:defcfun ("fomus_clear" fomus_clear) :void
  (f :pointer))

(cffi:defcfun ("fomus_get_ival" fomus_get_ival) :long
  (f :pointer)
  (set :string))

(cffi:defcfun ("fomus_get_rval" fomus_get_rval) fomus_rat
  (f :pointer)
  (set :string))

(cffi:defcfun ("fomus_get_fval" fomus_get_fval) :double
  (f :pointer)
  (set :string))

(cffi:defcfun ("fomus_get_sval" fomus_get_sval) :string
  (f :pointer)
  (set :string))

(cffi:defcfun ("fomus_ival" fomus_ival) :void
  (f :pointer)
  (par :int)
  (act :int)
  (val :long))

(cffi:defcfun ("fomus_rval" fomus_rval) :void
  (f :pointer)
  (par :int)
  (act :int)
  (num :long)
  (den :long))

(cffi:defcfun ("fomus_mval" fomus_mval) :void
  (f :pointer)
  (par :int)
  (act :int)
  (val :long)
  (num :long)
  (den :long))

(cffi:defcfun ("fomus_fval" fomus_fval) :void
  (f :pointer)
  (par :int)
  (act :int)
  (val :double))

(cffi:defcfun ("fomus_sval" fomus_sval) :void
  (f :pointer)
  (par :int)
  (act :int)
  (str :string))

(cffi:defcfun ("fomus_act" fomus_act) :void
  (f :pointer)
  (par :int)
  (act :int))

(cffi:defcfun ("fomus_rt" fomus_rt) :void)

(cffi:defcfun ("fomus_ivalrt" fomus_ivalrt) :void
  (f :pointer)
  (par :int)
  (act :int)
  (val :long))

(cffi:defcfun ("fomus_rvalrt" fomus_rvalrt) :void
  (f :pointer)
  (par :int)
  (act :int)
  (num :long)
  (den :long))

(cffi:defcfun ("fomus_mvalrt" fomus_mvalrt) :void
  (f :pointer)
  (par :int)
  (act :int)
  (val :long)
  (num :long)
  (den :long))

(cffi:defcfun ("fomus_fvalrt" fomus_fvalrt) :void
  (f :pointer)
  (par :int)
  (act :int)
  (val :double))

(cffi:defcfun ("fomus_svalrt" fomus_svalrt) :void
  (f :pointer)
  (par :int)
  (act :int)
  (str :string))

(cffi:defcfun ("fomus_actrt" fomus_actrt) :void
  (f :pointer)
  (par :int)
  (act :int))

(cffi:defcfun ("fomus_load" fomus_load) :void
  (f :pointer)
  (filename :string))

(cffi:defcfun ("fomus_parse" fomus_parse) :void
  (f :pointer)
  (input :string))

(cffi:defcfun ("fomus_run" fomus_run) :void
  (f :pointer))

(cffi:defcfun ("fomus_copy" fomus_copy) :pointer
  (f :pointer))

(cffi:defcfun ("fomus_save" fomus_save) :void
  (f :pointer)
  (filename :string))

(cffi:defcfun ("fomus_set_outputs" fomus_set_outputs) :void
  (out :pointer)
  (err :pointer)
  (newline :int))


;; ------------------------------------------------------------------------------------------------------------------------
;; low-level stuff

(defvar *init* nil) ; initialized?
(defvar *fomus* nil) ; pointer to fomus instance

(defmacro errwrap (&body forms) ; right now FOMUS prints errors to STDERR--FIXME
  `(unwind-protect
	(progn ,@forms)
     (when (/= fomus_err 0) (error "Error in FOMUS"))))

(cffi:defcallback output :void ((str :string))
  (format t ";; ~A~%" str))

(defun init ()
  (unless *init*
    (fomus_set_outputs (cffi:callback output) (cffi:callback output) 0)
    (errwrap (fomus_init)) 
    (setf *init* t)))
(defun new ()
  ;; lib should get loaded here
  (init)
  (unless *fomus* (setf *fomus* (errwrap (fomus_new)))))

(defun filter (sets &rest keys) ; filter out keywords
  (loop for (a b) on sets by #'cddr
       unless (member a keys) collect a and collect b))

;; structs (these can be defined as global objects or can be embedded in a parent struct)
(defstruct (fmsmeasdef (:constructor make-measdef-aux))
  id sets)
(defun make-measdef (&rest sets &key id &allow-other-keys)
  (make-measdef-aux :id id :sets (filter sets :id)))

(defstruct (fmspart (:constructor make-part-aux))
  id inst sets)
(defun make-part (&rest sets &key id inst &allow-other-keys)
  (make-part-aux :id id :inst (if (consp inst) (apply #'make-inst inst) inst) :sets (filter sets :id :inst)))

(defstruct (fmsmetapart (:constructor make-metapart-aux))
  id parts sets)
(defun make-metapart (&rest sets &key id parts &allow-other-keys)
  (make-metapart-aux :id id :parts (mapcar (lambda (x) (if (listp x) (apply #'make-part x) x)) parts) :sets (filter sets :id :parts)))

(defstruct (fmspartsref (:constructor make-partsref-aux))
  id parts) ; sets is supposed to be missing
(defun make-partsref (&key id parts)
  (make-partsref-aux :id id :parts (mapcar (lambda (x) (if (listp x) (apply #'make-part x) x)) parts)))

(defstruct (fmsinst (:constructor make-inst-aux))
  template id staves imports export percinsts sets)
(defun make-inst (&rest sets &key template id staves imports export percinsts &allow-other-keys)
  (make-inst-aux :template template :id id :staves (mapcar (lambda (x) (apply #'make-staff x)) staves)
	 :imports (mapcar (lambda (x) (apply #'make-import x)) imports) :export (if (consp export) (apply #'make-export export) export)
	 :percinsts (mapcar (lambda (x) (if (listp x) (apply #'make-percinst x) x)) percinsts)
	 :sets (filter sets :template :id :staves :imports :export :percinsts)))

(defstruct (fmspercinst (:constructor make-percinst-aux))
  template id imports export sets)
(defun make-percinst (&rest sets &key template id imports export &allow-other-keys)
  (make-percinst-aux :template template :id id :imports (mapcar (lambda (x) (apply #'make-import x)) imports)
	 :export (if (consp export) (apply #'make-export export) export) :sets sets))

;; embedded structs (these can only exist inside a parent struct above)
(defstruct (fmspartmap (:constructor make-partmap-aux))
  part metapart sets)
(defun make-partmap (&rest sets &key part metapart &allow-other-keys)
  (make-inst-aux :part (if (listp part) (apply #'make-part part) part) :metapart (if (listp metapart) (apply #'make-metapart metapart) metapart)
	 :sets (filter sets :part :metapart)))

(defstruct (fmsclef (:constructor make-clef-aux))
  sets)
(defun make-clef (&rest sets &key &allow-other-keys)
  (make-inst-aux :sets sets))

(defstruct (fmsstaff (:constructor make-staff-aux))
  clefs sets)
(defun make-staff (&rest sets &key clefs &allow-other-keys)
  (make-inst-aux :clefs (mapcar (lambda (x) (apply #'make-clef x)) clefs) :sets (filter sets :clefs)))

(defstruct (fmsimport (:constructor make-import-aux))
  sets)
(defun make-import (&rest sets &key &allow-other-keys)
  (make-inst-aux :sets sets))

(defstruct (fmsexport (:constructor make-export-aux))
  sets)
(defun make-export (&rest sets &key &allow-other-keys)
  (make-inst-aux :sets sets))

;; the main send-data-to-fomus function
(defun send (par act &optional val)
  (flet ((sendmult (lpar lact lval) ; if a list, "send" everything IN the list without actually sending a list
	   (loop for e in (if (listp lval) lval (list lval)) do (send lpar lact e)))
	 (sendstr (strval valpar)
	   (loop for (a b) on strval by #'cddr
	      do
		(send (cffi:foreign-enum-value 'fomus_param :fomus_par_setting) (cffi:foreign-enum-value 'fomus_action :fomus_act_set) a)
		(send valpar (cffi:foreign-enum-value 'fomus_action :fomus_act_set) (case b ((nil) 'no) ((t) 'yes) (otherwise b))))))
    (etypecase val
      (null (errwrap (fomus_act *fomus* par act)))
      (integer (errwrap (fomus_ival *fomus* par act val)))
      (rational (errwrap (fomus_rval *fomus* par act (numerator val) (denominator val))))
      (float (errwrap (fomus_fval *fomus* par act val)))
      (symbol (errwrap (fomus_sval *fomus* par act (string-downcase val))))
      (string (errwrap (fomus_sval *fomus* par act val)))
      (pathname (errwrap (fomus_sval *fomus* par act (namestring (translate-logical-pathname val)))))
      (list (loop
	       initially (errwrap (fomus_act *fomus* (cffi:foreign-enum-value 'fomus_param :fomus_par_list) (cffi:foreign-enum-value 'fomus_action :fomus_act_start)))
	       for e in val
	       do (send (cffi:foreign-enum-value 'fomus_param :fomus_par_list) (cffi:foreign-enum-value 'fomus_action :fomus_act_add) e)
	       finally
		 (errwrap (fomus_act *fomus* (cffi:foreign-enum-value 'fomus_param :fomus_par_list) (cffi:foreign-enum-value 'fomus_action :fomus_act_end)))
		 (errwrap (fomus_act *fomus* par act))))
      (fmsmeasdef ; ids, names, and other things are sent separately because they don't behave like normal settings
       (when (fmsmeasdef-id val) (send (cffi:foreign-enum-value 'fomus_param :fomus_par_measdef_id) (cffi:foreign-enum-value 'fomus_action :fomus_act_set) (fmsmeasdef-id val)))
       (sendstr (fmsmeasdef-sets val) (cffi:foreign-enum-value 'fomus_param :fomus_par_measdef_settingval))
       (errwrap (fomus_act *fomus* par act)))
      (fmspart
       (when (fmspart-id val) (send (cffi:foreign-enum-value 'fomus_param :fomus_par_part_id) (cffi:foreign-enum-value 'fomus_action :fomus_act_set) (fmspart-id val)))
       (when (fmspart-inst val) (send (cffi:foreign-enum-value 'fomus_param :fomus_par_part_inst) (cffi:foreign-enum-value 'fomus_action :fomus_act_set) (fmspart-inst val)))
       (sendstr (fmspart-sets val) (cffi:foreign-enum-value 'fomus_param :fomus_par_part_settingval))
       (errwrap (fomus_act *fomus* par act)))
      (fmsmetapart
       (when (fmsmetapart-id val) (send (cffi:foreign-enum-value 'fomus_param :fomus_par_metapart_id) (cffi:foreign-enum-value 'fomus_action :fomus_act_set) (fmsmetapart-id val)))
       (when (fmsmetapart-parts val) (sendmult (cffi:foreign-enum-value 'fomus_param :fomus_par_metapart_parts) (cffi:foreign-enum-value 'fomus_action :fomus_act_set) (fmsmetapart-parts val)))
       (sendstr (fmsmetapart-sets val) (cffi:foreign-enum-value 'fomus_param :fomus_par_metapart_settingval))
       (errwrap (fomus_act *fomus* par act)))
      (fmspartsref
       (when (fmspartsref-id val) (send (cffi:foreign-enum-value 'fomus_param :fomus_par_partsref_id) (cffi:foreign-enum-value 'fomus_action :fomus_act_set) (fmspartsref-id val)))
       (when (fmspartsref-parts val) (sendmult (cffi:foreign-enum-value 'fomus_param :fomus_par_partsref_parts) (cffi:foreign-enum-value 'fomus_action :fomus_act_set) (fmspartsref-parts val)))
       (errwrap (fomus_act *fomus* par act))) ; sendstr is supposed to be missing from this one
      (fmsinst
       (when (fmsinst-template val) (send (cffi:foreign-enum-value 'fomus_param :fomus_par_inst_template) (cffi:foreign-enum-value 'fomus_action :fomus_act_set) (fmsinst-template val)))
       (when (fmsinst-id val) (send (cffi:foreign-enum-value 'fomus_param :fomus_par_inst_id) (cffi:foreign-enum-value 'fomus_action :fomus_act_set) (fmsinst-id val)))
       (when (fmsinst-staves val) (sendmult (cffi:foreign-enum-value 'fomus_param :fomus_par_inst_staves) (cffi:foreign-enum-value 'fomus_action :fomus_act_set) (fmsinst-staves val)))
       (when (fmsinst-imports val) (sendmult (cffi:foreign-enum-value 'fomus_param :fomus_par_inst_imports) (cffi:foreign-enum-value 'fomus_action :fomus_act_set) (fmsinst-imports val)))
       (when (fmsinst-export val) (send (cffi:foreign-enum-value 'fomus_param :fomus_par_inst_export) (cffi:foreign-enum-value 'fomus_action :fomus_act_set) (fmsinst-export val)))
       (when (fmsinst-percinsts val) (sendmult (cffi:foreign-enum-value 'fomus_param :fomus_par_inst_percinsts) (cffi:foreign-enum-value 'fomus_action :fomus_act_set) (fmsinst-percinsts val)))
       (sendstr (fmsinst-sets val) (cffi:foreign-enum-value 'fomus_param :fomus_par_inst_settingval))
       (errwrap (fomus_act *fomus* par act)))
      (fmspercinst
       (when (fmspercinst-template val) (send (cffi:foreign-enum-value 'fomus_param :fomus_par_percinst_template) (cffi:foreign-enum-value 'fomus_action :fomus_act_set) (fmspercinst-template val)))
       (when (fmspercinst-id val) (send (cffi:foreign-enum-value 'fomus_param :fomus_par_percinst_id) (cffi:foreign-enum-value 'fomus_action :fomus_act_set) (fmspercinst-id val)))
       (when (fmspercinst-imports val) (sendmult (cffi:foreign-enum-value 'fomus_param :fomus_par_percinst_imports) (cffi:foreign-enum-value 'fomus_action :fomus_act_set) (fmspercinst-imports val)))
       (when (fmspercinst-export val) (send (cffi:foreign-enum-value 'fomus_param :fomus_par_percinst_export) (cffi:foreign-enum-value 'fomus_action :fomus_act_set) (fmspercinst-export val)))
       (sendstr (fmspercinst-sets val) (cffi:foreign-enum-value 'fomus_param :fomus_par_percinst_settingval))
       (errwrap (fomus_act *fomus* par act)))
      (fmspartmap
       (when (fmspartmap-part val) (send (cffi:foreign-enum-value 'fomus_param :fomus_par_partmap_part) (cffi:foreign-enum-value 'fomus_action :fomus_act_set) (fmspartmap-part val)))
       (when (fmspartmap-metapart val) (send (cffi:foreign-enum-value 'fomus_param :fomus_par_partmap_metapart) (cffi:foreign-enum-value 'fomus_action :fomus_act_set) (fmspartmap-metapart val)))
       (sendstr (fmspartmap-sets val) (cffi:foreign-enum-value 'fomus_param :fomus_par_partmap_settingval))
       (errwrap (fomus_act *fomus* par act)))
      (fmsclef
       (sendstr (fmsclef-sets val) (cffi:foreign-enum-value 'fomus_param :fomus_par_clef_settingval))
       (errwrap (fomus_act *fomus* par act)))
      (fmsstaff
       (when (fmsstaff-clefs val) (sendmult (cffi:foreign-enum-value 'fomus_param :fomus_par_staff_clefs) (cffi:foreign-enum-value 'fomus_action :fomus_act_set) (fmsstaff-clefs val)))
       (sendstr (fmsstaff-sets val) (cffi:foreign-enum-value 'fomus_param :fomus_par_staff_settingval))
       (errwrap (fomus_act *fomus* par act)))
      (fmsimport
       (sendstr (fmsimport-sets val) (cffi:foreign-enum-value 'fomus_param :fomus_par_import_settingval))
       (errwrap (fomus_act *fomus* par act)))
      (fmsexport
       (sendstr (fmsexport-sets val) (cffi:foreign-enum-value 'fomus_param :fomus_par_export_settingval))
       (errwrap (fomus_act *fomus* par act))))))

;; ------------------------------------------------------------------------------------------------------------------------
;; aux

(defun entry (time dur part voice grtime marks settings)
  (when time (send (cffi:foreign-enum-value 'fomus_param :fomus_par_time) (cffi:foreign-enum-value 'fomus_action :fomus_act_set) time))
  (when dur (send (cffi:foreign-enum-value 'fomus_param :fomus_par_duration) (cffi:foreign-enum-value 'fomus_action :fomus_act_set) dur))
  (when part (send (cffi:foreign-enum-value 'fomus_param :fomus_par_part) (cffi:foreign-enum-value 'fomus_action :fomus_act_set) part))
  (when voice (send (cffi:foreign-enum-value 'fomus_param :fomus_par_voice) (cffi:foreign-enum-value 'fomus_action :fomus_act_set) voice))
  (when grtime (send (cffi:foreign-enum-value 'fomus_param :fomus_par_gracetime) (cffi:foreign-enum-value 'fomus_action :fomus_act_set) grtime))
  (loop for e in marks
     do (destructuring-bind (mark &optional arg1 arg2) (if (listp e) e (list e))
	  (send (cffi:foreign-enum-value 'fomus_param :fomus_par_mark) (cffi:foreign-enum-value 'fomus_action :fomus_act_set) mark)
	  (when arg1 (send (cffi:foreign-enum-value 'fomus_param :fomus_par_markval) (cffi:foreign-enum-value 'fomus_action :fomus_act_add) arg1))
	  (when arg2 (send (cffi:foreign-enum-value 'fomus_param :fomus_par_markval) (cffi:foreign-enum-value 'fomus_action :fomus_act_add) arg2))
	  (send (cffi:foreign-enum-value 'fomus_param :fomus_par_marklist) (cffi:foreign-enum-value 'fomus_action :fomus_act_add))))
  (loop for e in settings
       do (destructuring-bind (set val) e
	    (send (cffi:foreign-enum-value 'fomus_param :fomus_par_setting) (cffi:foreign-enum-value 'fomus_action :fomus_act_set) set)
	    (send (cffi:foreign-enum-value 'fomus_param :fomus_par_note_settingval) (cffi:foreign-enum-value 'fomus_action :fomus_act_add) val))))

;; ------------------------------------------------------------------------------------------------------------------------
;; user functions

(defun fomus-load (filename)
  (new)
  (errwrap (fomus_load *fomus* (namestring (translate-logical-pathname filename)))))

;; invalidates the instance, so set it to nil
(defun fomus-run ()
  (new)
  (unwind-protect (errwrap (fomus_run *fomus*)) (setf *fomus* nil)))

;; (defun fomus-free ()
;;   (unless *init* (errwrap (fomus_init)) (setf *init* t))
;;   (when *fomus* (fomus_free *fomus*)))

(defun fomus-version ()
  (fomus_version))

;; event entry
(defun fomus-note (&key time dur pitch part voice dyn grtime marks settings)
  (new)
  (when pitch (send (cffi:foreign-enum-value 'fomus_param :fomus_par_pitch) (cffi:foreign-enum-value 'fomus_action :fomus_act_set) pitch))
  (when dyn (send (cffi:foreign-enum-value 'fomus_param :fomus_par_dynlevel) (cffi:foreign-enum-value 'fomus_action :fomus_act_set) dyn))
  (entry time dur part voice grtime marks settings)
  (send (cffi:foreign-enum-value 'fomus_param :fomus_par_noteevent) (cffi:foreign-enum-value 'fomus_action :fomus_act_add)))

(defun fomus-rest (&key time dur part voice grtime marks settings)
  (new)
  (entry time dur part voice grtime marks settings)
  (send (cffi:foreign-enum-value 'fomus_param :fomus_par_restevent) (cffi:foreign-enum-value 'fomus_action :fomus_act_add)))

(defun fomus-mark (&key time dur part voice grtime marks settings)
  (new)
  (entry time dur part voice grtime marks settings)
  (send (cffi:foreign-enum-value 'fomus_param :fomus_par_markevent) (cffi:foreign-enum-value 'fomus_action :fomus_act_add)))

(defun fomus-meas (&key time dur settings)
  (new)
  (when time (send (cffi:foreign-enum-value 'fomus_param :fomus_par_time) (cffi:foreign-enum-value 'fomus_action :fomus_act_set) time))
  (when dur (send (cffi:foreign-enum-value 'fomus_param :fomus_par_duration) (cffi:foreign-enum-value 'fomus_action :fomus_act_set) dur))
  (when settings (send (cffi:foreign-enum-value 'fomus_param :fomus_par_measdef) (cffi:foreign-enum-value 'fomus_action :fomus_act_add) (apply #'make-measdef settings)))
  (send (cffi:foreign-enum-value 'fomus_param :fomus_par_meas) (cffi:foreign-enum-value 'fomus_action :fomus_act_add)))

;; functions for defining global structs
(defun fomus-measdef (&rest settings)
  (new)
  (send (cffi:foreign-enum-value 'fomus_param :fomus_par_measdef) (cffi:foreign-enum-value 'fomus_action :fomus_act_add) (apply #'make-measdef settings)))

(defun fomus-part (&rest settings)
  (new)
  (send (cffi:foreign-enum-value 'fomus_param :fomus_par_part) (cffi:foreign-enum-value 'fomus_action :fomus_act_add) (apply #'make-part settings)))

(defun fomus-metapart (&rest settings)
  (new)
  (send (cffi:foreign-enum-value 'fomus_param :fomus_par_metapart) (cffi:foreign-enum-value 'fomus_action :fomus_act_add) (apply #'make-metapart settings)))

(defun fomus-partsref (&rest settings)
  (new)
  (send (cffi:foreign-enum-value 'fomus_param :fomus_par_partsref) (cffi:foreign-enum-value 'fomus_action :fomus_act_add) (apply #'make-partsref settings)))

(defun fomus-inst (&rest settings)
  (new)
  (send (cffi:foreign-enum-value 'fomus_param :fomus_par_inst) (cffi:foreign-enum-value 'fomus_action :fomus_act_add) (apply #'make-inst settings)))

(defun fomus-percinst (&rest settings)
  (new)
  (send (cffi:foreign-enum-value 'fomus_param :fomus_par_percinst) (cffi:foreign-enum-value 'fomus_action :fomus_act_add) (apply #'make-percinst settings)))

;; global settings
(defun fomus-setting (setting val &optional append)
  (new)
  (send (cffi:foreign-enum-value 'fomus_param :fomus_par_setting) (cffi:foreign-enum-value 'fomus_action :fomus_act_set) setting)
  (send (cffi:foreign-enum-value 'fomus_param :fomus_par_settingval)
	(if append (cffi:foreign-enum-value 'fomus_action :fomus_act_append) (cffi:foreign-enum-value 'fomus_action :fomus_act_set))
	(case val ((nil) 'no) ((t) 'yes) (otherwise val))))

;; with-score
(defmacro with-fomus-score (&body forms)
  `(progn
     (init)
     (let ((*fomus* (errwrap (fomus_new))))
       (unwind-protect
	    (progn ,@forms)
	 (when *fomus* (fomus_free *fomus*))))))
